#include "ddk.h"
#include "gbadef.h"
#include "krnobj.h"	

/*
 * global function
 */
GLOBAL_FUNCTION( KeUpdateTickCount )
GLOBAL_FUNCTION( KeInitializeDispatcherHeader )

/*
 * external functions
 */
extern KERNELAPI BOOL TmWaitForObject(
	IN	PDISPATCHER_HEADER				pDispatcherHeader,
	IN	ULONG							Timeout
	);
extern KERNELAPI KE_WAIT_STATUS TmQueryWaitStatus(VOID);
extern KERNELAPI VOID TmWaitScheduler(VOID);

/*
 * initialization entry points
 */
extern KERNELAPI BOOL HalInitializeModule(VOID);
extern KERNELAPI BOOL MmInitializeModule(VOID);
extern KERNELAPI BOOL TmInitializeModule(VOID);

/*
 * variables
 */
static ULONG		m_KernelTickCount;


//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
VOID
KeInitializeDispatcherHeader(
	OUT	PDISPATCHER_HEADER				pDispatcherHeader,
	IN	ULONG							Signature,
	IN	ULONG							SignalState
	)
{
	// check parameter
	if( !pDispatcherHeader ) { return; }
	
	// init
	pDispatcherHeader->signature	= Signature;
	pDispatcherHeader->signal_state	= SignalState;
	RtlInitializeList( &pDispatcherHeader->waitobj_list );
	RtlInitializeList( &pDispatcherHeader->wait_list );
}

//////////////////////////////////////////////////////////////////////////////////////////
static
BOOL
KepIsSupportedObjectType(
	IN	ULONG							Signature
	)
{
	switch( Signature )
	{
	case TCB_SIGNATURE:
	case KMUTEX_SIGNATURE:
	case KSEMAPHORE_SIGNATURE:
	case KEVENT_SIGNATURE:
	case KTIMER_SIGNATURE:
		{
			return TRUE;
		}
	}
	
	return FALSE;
}

static
BOOL
KepTrySingleObject(
	IN	PVOID							pObject
	)
{
	PDISPATCHER_HEADER		pt_header = NULL;
	BOOL					is_signaled;
	BOOL					use_int;
	
	// check use_int
	use_int = IS_INT_IME_ENABLED();
	
	// get dispatcher header
	pt_header = (PDISPATCHER_HEADER)pObject;
	
	// disable interrupts
	if( use_int ) { DISABLE_INTERRUPTS(); }

	// get default signal state
	is_signaled = (pt_header->signal_state ? TRUE : FALSE);
	
	// check type
	switch( pt_header->signature )
	{
	case TCB_SIGNATURE:
		{
			;
		}
		break;
	
	case KMUTEX_SIGNATURE:
		{
			if( pt_header->signal_state )
			{
				pt_header->signal_state	= 0;
			}
		}
		break;
		
	case KSEMAPHORE_SIGNATURE:
		{
			PKSEMAPHORE		pt_semaphore = (PKSEMAPHORE)pObject;
			
			if( pt_semaphore->count < pt_semaphore->limit )
			{
				pt_semaphore->count++;
				is_signaled = TRUE;
				if( pt_semaphore->count >= pt_semaphore->limit )
				{
					pt_header->signal_state = 0;
				}
			}
		}
		break;
		
	case KEVENT_SIGNATURE:
		{
			PKEVENT		pt_event = (PKEVENT)pObject;
			
			if( pt_header->signal_state && pt_event->type == SynchronizationEvent )
			{
				pt_header->signal_state = 0;
			}
		}
		break;
		
	case KTIMER_SIGNATURE:
		{
			PKTIMER		pt_timer = (PKTIMER)pObject;
			
			// is timeouted?
			if( pt_timer->timeout <= KeQueryTickCount() )
			{
				pt_timer->header.signal_state = 1;
				is_signaled = TRUE;
			}
		}
		break;
		
	default:
		return FALSE;
	}
	
	// enable interrupts
	if( use_int ) { ENABLE_INTERRUPTS(); }
	
	return is_signaled;
}

KERNELAPI
KE_WAIT_STATUS
KeWaitForSingleObject(
	IN	PVOID							pObject,
	IN	ULONG							Timeout // milli seconds
	)
{
	KE_WAIT_STATUS			wait_result = KE_WAIT_FAILED;
	PDISPATCHER_HEADER		pt_header = NULL;
	BOOL					is_signaled;
	
	// check parameter
	if( !pObject ) { return KE_WAIT_FAILED; }
	
	// get dispatcher header
	pt_header = (PDISPATCHER_HEADER)pObject;
	
	// is supported type?
	if( !KepIsSupportedObjectType(pt_header->signature) ) { return KE_WAIT_FAILED; }
	
	// try!
	is_signaled = KepTrySingleObject(pObject);
	if( is_signaled )
	{
		wait_result = KE_WAIT_SUCCESS;
	}
	else
	{
		if( Timeout )
		{
			// wait!
			TmWaitForObject( pt_header, Timeout );
			wait_result = TmQueryWaitStatus();
		}
		else
		{
			wait_result = KE_WAIT_TIMEOUT;
		}
	}

	return wait_result;
}

//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
VOID
KeInitializeMutex(
	IN	PKMUTEX							pMutexObj
	)
{
	// check parameter
	if( !pMutexObj ) { return; }
	
	// init
	memset( pMutexObj, 0, sizeof(KMUTEX) );
	KeInitializeDispatcherHeader( &pMutexObj->header, KMUTEX_SIGNATURE, 1 );
}

KERNELAPI
BOOL
KeReadStateMutex(
	IN	PKMUTEX							pMutexObj,
	OUT	PULONG							pSignalState
	)
{
	// check parameter
	if( !pMutexObj || !pSignalState ) { return FALSE; }
	if( pMutexObj->header.signature != KMUTEX_SIGNATURE ) { return FALSE; }
	
	// set state
	*pSignalState = pMutexObj->header.signal_state;
	
	return TRUE;
}
	
KERNELAPI
VOID
KeReleaseMutex(
	IN	PKMUTEX							pMutexObj,
	IN	BOOL							bWait
	)
{
	// check parameter
	if( !pMutexObj ) { return; }
	if( pMutexObj->header.signature != KMUTEX_SIGNATURE ) { return; }

	// release!
	pMutexObj->header.signal_state = 1;
	
	// wait scheduler
	if( !bWait )
	{
		TmWaitScheduler();
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
VOID
KeInitializeSemaphore(
	IN	PKSEMAPHORE						pSemaphoreObj,
	IN	ULONG							Count,
	IN	ULONG							Limit
	)
{
	// check parameter
	if( !pSemaphoreObj || !Limit || (Count>Limit) ) { return; }
	
	// init
	memset( pSemaphoreObj, 0, sizeof(KSEMAPHORE) );
	KeInitializeDispatcherHeader(
		&pSemaphoreObj->header,
		KSEMAPHORE_SIGNATURE,
		(Count!=Limit && Count ? 1 : 0)
		);
	pSemaphoreObj->count = Count;
	pSemaphoreObj->limit = Limit;
}

KERNELAPI
BOOL
KeReadStateSemaphore(
	IN	PKSEMAPHORE						pSemaphoreObj,
	OUT	PULONG							pSignalState
	)
{
	// check parameter
	if( !pSemaphoreObj || !pSignalState ) { return FALSE; }
	if( pSemaphoreObj->header.signature != KSEMAPHORE_SIGNATURE ) { return FALSE; }
	
	// set state
	*pSignalState = pSemaphoreObj->header.signal_state;
	
	return TRUE;
}
	
KERNELAPI
VOID
KeReleaseSemaphore(
	IN	PKSEMAPHORE						pSemaphoreObj,
	IN	BOOL							bWait
	)
{
	BOOL	use_int;
	
	// check parameter
	if( !pSemaphoreObj ) { return; }
	if( pSemaphoreObj->header.signature != KSEMAPHORE_SIGNATURE ) { return; }
	if( pSemaphoreObj->count==0 ) { return; }
	
	// check use_int
	use_int = IS_INT_IME_ENABLED();
	
	// disable interrupts
	if( use_int ) { DISABLE_INTERRUPTS(); }

	// release!
	pSemaphoreObj->count--;
	pSemaphoreObj->header.signal_state = 1;
	
	// enable interrupts
	if( use_int ) { ENABLE_INTERRUPTS(); }

	// wait scheduler
	if( !bWait )
	{
		TmWaitScheduler();
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
VOID
KeInitializeEvent(
	IN	PKEVENT							pEventObj,
	IN	KEVENT_TYPE						EventType,
	IN	BOOL							State
	)
{
	// check parameter
	if( !pEventObj ) { return; }
	
	// init
	memset( pEventObj, 0, sizeof(KEVENT) );
	KeInitializeDispatcherHeader( &pEventObj->header, KEVENT_SIGNATURE, (State?1:0) );
	pEventObj->type = EventType;
}
	
KERNELAPI
BOOL
KeReadStateEvent(
	IN	PKEVENT							pEventObj,
	OUT	PULONG							pSignalState
	)
{
	// check parameter
	if( !pEventObj || !pSignalState ) { return FALSE; }
	if( pEventObj->header.signature != KEVENT_SIGNATURE ) { return FALSE; }
	
	// set state
	*pSignalState = pEventObj->header.signal_state;
	
	return TRUE;
}
	
KERNELAPI
VOID
KeClearEvent(
	IN	PKEVENT							pEventObj,
	IN	BOOL							bWait
	)
{
	// check parameter
	if( !pEventObj ) { return; }
	if( pEventObj->header.signature != KEVENT_SIGNATURE ) { return; }

	// clear!
	pEventObj->header.signal_state = 0;
	
	// wait scheduler
	if( !bWait )
	{
		TmWaitScheduler();
	}
}

KERNELAPI
VOID
KeSetEvent(
	IN	PKEVENT							pEventObj,
	IN	BOOL							bWait
	)
{
	// check parameter
	if( !pEventObj ) { return; }
	if( pEventObj->header.signature != KEVENT_SIGNATURE ) { return; }

	// set!
	pEventObj->header.signal_state = 1;
	
	// wait scheduler
	if( !bWait )
	{
		TmWaitScheduler();
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
VOID
KeInitializeTimer(
	IN	PKTIMER							pTimerObj
	)
{
	// check parameter
	if( !pTimerObj ) { return; }
	
	// init
	memset( pTimerObj, 0, sizeof(KTIMER) );
	KeInitializeDispatcherHeader( &pTimerObj->header, KTIMER_SIGNATURE, 0 );
}

KERNELAPI
BOOL
KeSetTimer(
	IN	PKTIMER							pTimerObj,
	IN	ULONG							Timeout // milli seconds
	)
{
	// check parameter
	if( !pTimerObj ) { return FALSE; }
	if( pTimerObj->header.signature != KTIMER_SIGNATURE ) { return FALSE; }
	
	// set new timeout
	pTimerObj->timeout
		= (Timeout==INFINITE ? INFINITE : KeQueryTickCount()+Timeout);
	
	return TRUE;
}

KERNELAPI
BOOL
KeCancelTimer(
	IN	PKTIMER							pTimerObj
	)
{
	// check parameter
	if( !pTimerObj ) { return FALSE; }
	if( pTimerObj->header.signature != KTIMER_SIGNATURE ) { return FALSE; }
	
	// reset
	pTimerObj->timeout = 0;
	
	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
VOID
KeSleep(
	IN	ULONG							SleepTime // milli seconds
	)
{
	KTIMER		timer_obj;
	
	// check parameter
	if( SleepTime == 0 ) { return; }
	
	// init obj
	KeInitializeTimer( &timer_obj );
	KeSetTimer( &timer_obj, SleepTime );
	
	// wait!
	KeWaitForSingleObject( (PVOID)&timer_obj, SleepTime );
}

KERNELAPI
VOID
KeUpdateTickCount(
	IN	ULONG							TickCount
	)
{
	m_KernelTickCount += TickCount;
}

KERNELAPI
ULONG
KeQueryTickCount(VOID)
{
	return m_KernelTickCount;
}

//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
BOOL
KrnInitializeModule(VOID)
{
	BOOL	result = TRUE;

	// reset kernel tickcount
	m_KernelTickCount = 0;
    
    // initialize HAL
    result = HalInitializeModule();
    if( !result )
    {
        goto $exit;
    }

	// initialize Memory Manager
	result = MmInitializeModule();
	if( !result )
	{
	    goto $exit;
 	}
 	
 	// initialize Task Manager
 	result = TmInitializeModule();
 	if( !result )
 	{
		goto $exit;
	}
 	
$exit:
    return result;
}


