#include "ddk.h"

/*
 $$ gameboy advance memory map! $$
 ---------------------------------
 + system rom________ : 0x00000000 ~ 0x00003fff     (16 kbytes, 32bit wide)
 + external work ram_ : 0x02000000 ~ 0x0203ffff     (256 kbytes, 16bit wide)
 + internal work ram_ : 0x03000000 ~ 0x03007fff     (32 kbytes, 32bit wide)
 + io ram ___________ : 0x04000000 ~ 0x040003ff     (1 kbytes, 32bit wide)
 + palette ram_______ : 0x05000000 ~ 0x050003ff     (1 kbytes, 16bit wide)
 + vram______________ : 0x06000000 ~ 0x06017fff     (96 kbytes, 16bit wide)
 + oam_______________ : 0x07000000 ~ 0x070003ff     (1 kbytes, 32bit wide)
 + game pak rom______ : 0x08000000 ~ 0x09ffffff     (max : 32 mbytes, 16bit wide)
 + game pak ram______ : 0x0e000000 ~ 0x0e00ffff     (max : 64 kbytes, 8bit wide)
 */
 
 /*
 $$ internal work ram map $$
 ---------------------------
 + kernel global variables_ : 0x03000000 ~ 0x03001fff		(8 kbytes)
 + unknown exception stack_ : 0x03002000 ~ 0x03002fff       (4 kbytes)
 + SVC mode stack__________ : 0x03003000 ~ 0x03004fff		(8 kbytes)
 + IRQ mode stack__________ : 0x03005000 ~ 0x03006fff		(8 kbytes)
 + memory pool map_________ : 0x03007000 ~ 0x030073ff		(1 kbytes)
 + GBA special use_________ : 0x03007400 ~ 0x03007fff		(3 kbytes)
 */
 
 /*
 $$ external work ram map $$
 ---------------------------
 + 8192 memory blocks (32bytes per block)
 */
 
/*
 * definition
 */
#define MM_BD_SIGNATURE			'mmbd'

#define MM_MEM_POOL_MAP_ADDR	0x03007000
#define MM_MEM_POOL_ADDR		0x02000000
#define MM_MEM_POOL_SIZE		(256*1024)

#define MM_BYTES_PER_BLOCK		32
#define MM_NUM_OF_BLOCKS		(MM_MEM_POOL_SIZE / MM_BYTES_PER_BLOCK)

/*
 * macro function
 */
#define LOCK_MEMORY_MANAGER()		DISABLE_INTERRUPTS()
#define UNLOCK_MEMORY_MANAGER()		ENABLE_INTERRUPTS()

/*
 * macro function (lsb=first, msb=last)
 */
#define MM_IS_FREE_BLOCK(n)												\
	( ( *((PUCHAR)MM_MEM_POOL_MAP_ADDR + ((n)/8)) & (1 << ((n)%8)) ) == 0 )

#define MM_MARK_FREE_BLOCK(n)											\
    ( *((PUCHAR)MM_MEM_POOL_MAP_ADDR + ((n)/8)) &= (~(1 << ((n)%8))) )

#define MM_MARK_USED_BLOCK(n)											\
	( *((PUCHAR)MM_MEM_POOL_MAP_ADDR + ((n)/8)) |= (1 << ((n)%8)) )

#define MM_GET_POOL_ADDR(n)												\
	( MM_MEM_POOL_ADDR + (MM_BYTES_PER_BLOCK*(n)) )

#define MM_GET_MAP_NUM(addr)											\
	( ((ULONG)(addr) - MM_MEM_POOL_ADDR) / MM_BYTES_PER_BLOCK )	// addr = substracted 16
 
/*
 * structures
 */
typedef struct _MM_BLOCK_DESC {
    ULONG						signature;		// 'mmbd'
    ULONG						num_of_blocks;	// with itself
} MM_BLOCK_DESC, *PMM_BLOCK_DESC;

//
typedef struct _MEMORY_MANAGER_BLOCK {
    //KMUTEX						mutex_obj;
	char a;
} MEMORY_MANAGER_BLOCK, *PMEMORY_MANAGER_BLOCK;

/*
 * variables
 */
static MEMORY_MANAGER_BLOCK			m_MemoryMgr;

//////////////////////////////////////////////////////////////////////////////////////////
static
BOOL
MmpInitMemPoolMap(VOID)
{
    PULONG		pt_pool_map = NULL;
    ULONG		num_of_tokens;
    ULONG		i;
    
    // set info
    pt_pool_map 	= (PULONG)MM_MEM_POOL_MAP_ADDR;
	num_of_tokens	= (MM_MEM_POOL_SIZE/MM_BYTES_PER_BLOCK/32); /* 32 = bits per ULONG */
	
	// init!
	for(i=0; i<num_of_tokens; i++)
	{
	    pt_pool_map[i] = 0;
	}
	
	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
PVOID
MmAllocate(
	IN	ULONG							NumOfBytes
	)
{
    PVOID		pt_mem = NULL;
    ULONG		num_of_blocks = 0;
	ULONG		free_blocks = 0;
	ULONG		start_block = 0;
    ULONG		i;
    
    // check parameter
    if( !NumOfBytes )
    {
        return NULL;
    }
    
    // getto!
    NumOfBytes += sizeof(MM_BLOCK_DESC);
    num_of_blocks = ((NumOfBytes + (MM_BYTES_PER_BLOCK-1)) / MM_BYTES_PER_BLOCK);
    
    // do
    LOCK_MEMORY_MANAGER();
    do
    {
	    // find
		free_blocks = 0;
		for(i=0; i<MM_NUM_OF_BLOCKS; i++)
		{
			// is all used block?
			if( *((PUCHAR)MM_MEM_POOL_MAP_ADDR + (i/8)) == 0xff )
			{
				i += 7;
				continue;
			}
	
			// find free block
			free_blocks = ( MM_IS_FREE_BLOCK(i) ? (free_blocks+1) : 0 );
	
			// find all?
			if( free_blocks >= num_of_blocks )
			{
				break;
			}
		}
		if( i == MM_NUM_OF_BLOCKS )
		{
		    break;
		}
	
		// set start block
		start_block = (i - (num_of_blocks-1));
	
		// mark all
		for(i=start_block; i<(start_block+num_of_blocks); i++)
		{
			MM_MARK_USED_BLOCK(i);
		}
	} while(FALSE);
	UNLOCK_MEMORY_MANAGER();
	
	// not enough memory?
	if( i == MM_NUM_OF_BLOCKS )
	{
	    return NULL;
	}

	// make block desc
	{
		PMM_BLOCK_DESC		pt_block_desc = NULL;

		// set
		pt_block_desc = (PMM_BLOCK_DESC)MM_GET_POOL_ADDR(start_block);
		pt_block_desc->signature		= MM_BD_SIGNATURE;
		pt_block_desc->num_of_blocks	= num_of_blocks;
	}

	// set free mem addr
	pt_mem = (PVOID)(MM_GET_POOL_ADDR(start_block) + sizeof(MM_BLOCK_DESC));
    
    return pt_mem;
}

KERNELAPI
VOID
MmFree(
	IN	PVOID							pMemAddr
	)
{
	PMM_BLOCK_DESC		pt_block_desc = NULL;
	ULONG				start_block;
	ULONG				i;

    // check parameter
	if( !pMemAddr )
	{
		return;
	}

	// get block desc
	pt_block_desc = (PMM_BLOCK_DESC)((ULONG)pMemAddr - sizeof(MM_BLOCK_DESC));

	// check signautre
	if( pt_block_desc->signature != MM_BD_SIGNATURE )
	{
		return;
	}

	// get start block
	start_block = MM_GET_MAP_NUM(pt_block_desc);
	
	// mark free!
	LOCK_MEMORY_MANAGER();
	for(i=0; i<pt_block_desc->num_of_blocks; i++)
	{
		MM_MARK_FREE_BLOCK( start_block + i );
	}
	UNLOCK_MEMORY_MANAGER();

	// nullify
	pt_block_desc->signature = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
BOOL
MmInitializeModule(VOID)
{
    BOOL	result = TRUE;
    
    // init memory manager
    memset( &m_MemoryMgr, 0, sizeof(m_MemoryMgr) );
    //KeInitializeMutex( &m_MemoryMgr.mutex_obj );
    
    // init memory pool map
    result = MmpInitMemPoolMap();
    if( !result )
    {
        goto $exit;
    }

	//
$exit:
    return result;
}
