#include "ddk.h"
#include "gbadef.h"
#include "krnobj.h"	

/*
 * global functions
 */
GLOBAL_FUNCTION( TmIrqScheduler )
GLOBAL_FUNCTION( TmStartMultiTasking )
GLOBAL_FUNCTION( TmWaitScheduler )

GLOBAL_FUNCTION( TmWaitForObject )
GLOBAL_FUNCTION( TmWakeupAllTasksWaitingObject );
GLOBAL_FUNCTION( TmWakeupFirstTaskWaitingObject );
GLOBAL_FUNCTION( TmQueryWaitStatus );

/*
 * external functions
 */
extern VOID TmpTaskContextSwitch(
	IN	ULONG							StackCurrent
 	);
extern KERNELAPI VOID KeInitializeDispatcherHeader(
	OUT	PDISPATCHER_HEADER				pDispatcherHeader,
	IN	ULONG							Signature,
	IN	ULONG							SignalState
	);
extern KERNELAPI VOID KeUpdateTickCount(
	IN	ULONG							TickCount
	);

/*
 * definition
 */
#define DEFAULT_STACK_SIZE					(2048-8)
#define CONTEXT_DATA_SIZE					(16*4)

#define PRIORITY_GROUP_COUNT				8

// 10 per sec -> 100  per sec = 1% overhead
// 10 per sec -> 1000 per sec = 7% overhead
#define MSECS_PER_TMR_INT					10 // use 100 per sec

/*
 * macro function
 */
#define CLEAR_TIMER_INTERRUPT_FLAG()		(GBA_INT_REG_IF |= GBA_INT_TIMER3)

//
#define PspGetTcbByHandle(handle)			(PTASK_CONTROL_BLOCK)(handle)
#define PspGetTcbByTaskList(pt_entry)		\
	(PTASK_CONTROL_BLOCK)((ULONG)pt_entry - (ULONG)(&((PTASK_CONTROL_BLOCK)0)->task_list))
#define PspGetTcbByReadyList(pt_entry)		\
	(PTASK_CONTROL_BLOCK)((ULONG)pt_entry - (ULONG)(&((PTASK_CONTROL_BLOCK)0)->ready_list))
#define PspGetTcbByWaitList(pt_entry)		\
	(PTASK_CONTROL_BLOCK)((ULONG)pt_entry - (ULONG)(&((PTASK_CONTROL_BLOCK)0)->wait_list))

/*
 * structure
 */
typedef ULONG						TASK_STATUS;
#define TASK_STATUS_SUSPENDED		0	// do not schedule or execute, when creating
#define TASK_STATUS_TERMINATED		1	// terminated, scheduler must delete it!
#define TASK_STATUS_READY			2	// ready to start
#define TASK_STATUS_WAITING			3	// waiting events
#define TASK_STATUS_RUNNING			4	// the running thread

typedef struct _TASK_CONTROL_BLOCK {
	DISPATCHER_HEADER				header;

	//
	LIST_ENTRY						task_list;	// task - task - task - ...
	LIST_ENTRY						ready_list;	// task - task - task - ...
	
	//
	LIST_ENTRY						wait_list;	// task - task - task - ...
	ULONG							wait_timeout; // end tick count
	KE_WAIT_STATUS					wait_status;

	// task info
	ULONG							tid;
	TASK_PRIORITY					priority;
	TASK_STATUS						status;
	
	// stack info
	ULONG							stack_size;
	ULONG							stack_start;
	ULONG							stack_cur;
} TASK_CONTROL_BLOCK, *PTASK_CONTROL_BLOCK;

//
typedef struct _TASK_MANAGER_BLOCK {
	ULONG							next_tid;

	//
	PTASK_CONTROL_BLOCK				pt_cur_task;

	//
	ULONG							task_count;
	LIST_ENTRY						task_list; // task - task - ....
	LIST_ENTRY						ready_list[PRIORITY_GROUP_COUNT]; // task - task - ....

	//
	LIST_ENTRY						waitobj_list; // obj - obj - ....
} TASK_MANAGER_BLOCK, *PTASK_MANAGER_BLOCK;

/*
 * variables
 */
static TASK_MANAGER_BLOCK			m_TaskMgr;


//////////////////////////////////////////////////////////////////////////////////////////
static
VOID
TmpTaskSelfDestroy(VOID)
{
	TmTerminateTask( TmGetCurrentTask() );
	while(1) ;
}

//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
VOID
TmStartMultiTasking(VOID)
{
	// init timer
	GBA_INT_REG_IE		|= GBA_INT_TIMER3;
	GBA_TMR_REG_TM3D	= 65372; // 100Hz
	GBA_TMR_REG_TM3CNT	= GBA_TMR_FREQ_1024 | GBA_TMR_OVERFLOW | GBA_TMR_ENABLE;
	
	// first task switichng
	m_TaskMgr.pt_cur_task->stack_cur += CONTEXT_DATA_SIZE;
	TmpTaskContextSwitch(
		m_TaskMgr.pt_cur_task->stack_cur - CONTEXT_DATA_SIZE
		);
}

KERNELAPI
VOID
TmWaitScheduler(VOID)
{
	ULONG	tick_count;
	
	tick_count = KeQueryTickCount();
	while( tick_count == KeQueryTickCount() ) ;
}

static
VOID
TmpWakeupTask(
	IN	PTASK_CONTROL_BLOCK				pTask,
	IN	KE_WAIT_STATUS					WaitStatus
	)
{
	// check parameter
	if( !pTask ) { return; }
	
	// 
	pTask->wait_status	= WaitStatus;
	pTask->status		= TASK_STATUS_READY;
	RtlInsertToListHead(
		&m_TaskMgr.ready_list[ (ULONG)pTask->priority ],
		&pTask->ready_list
		);

	// clear wait list
	RtlRemoveEntryFromList( &pTask->wait_list );
}

KERNELAPI
VOID
TmWakeupAllTasksWaitingObject(
	IN	PLIST_ENTRY						pWaitList,
	IN	KE_WAIT_STATUS					WaitStatus
	)
{
	PLIST_ENTRY		pt_entry = NULL;
	BOOL			use_int;
	
	// check parameter
	if( !pWaitList ) { return; }
	
	// check use_int
	use_int = IS_INT_IME_ENABLED();
	
	// disable interrupts
	if( use_int ) { DISABLE_INTERRUPTS(); }
	
	// wakeup all!
	for( pt_entry=pWaitList->Flink; pt_entry!=pWaitList; )
	{
		PTASK_CONTROL_BLOCK		pt_task = NULL;
		
		// get task block
		pt_task = PspGetTcbByWaitList( pt_entry );
		
		// wakeup!
		pt_entry = pt_entry->Flink;
		TmpWakeupTask( pt_task, WaitStatus );
	}
	
	// enable interrupts
	if( use_int ) { ENABLE_INTERRUPTS(); }
}

KERNELAPI
VOID
TmWakeupFirstTaskWaitingObject(
	IN	PLIST_ENTRY						pWaitList,
	IN	KE_WAIT_STATUS					WaitStatus
	)
{
	BOOL	use_int;
	
	// check parameter
	if( !pWaitList ) { return; }
	
	// check use_int
	use_int = IS_INT_IME_ENABLED();
	
	// disable interrupts
	if( use_int ) { DISABLE_INTERRUPTS(); }
	
	// wakeup!
	if( pWaitList->Flink != pWaitList )
	{
		PTASK_CONTROL_BLOCK		pt_task = NULL;
		
		// get task block
		pt_task = PspGetTcbByWaitList( pWaitList->Flink );
		
		// wakeup!
		TmpWakeupTask( pt_task, WaitStatus );
	}
	
	// enable interrupts
	if( use_int ) { ENABLE_INTERRUPTS(); }
}

KERNELAPI
BOOL
TmWaitForObject(
	IN	PDISPATCHER_HEADER				pDispatcherHeader,
	IN	ULONG							Timeout
	)
{
	PTASK_CONTROL_BLOCK		pt_task = m_TaskMgr.pt_cur_task;

	// check parameter
	if( !pDispatcherHeader ) { return FALSE; }

	// wait!
	DISABLE_INTERRUPTS();
	{
		// register waitobj entry
		// ! it'll be unregistered automately by TmpReleaseTasksWaitingObject() !
		if( RtlGetListCount( &pDispatcherHeader->waitobj_list ) == 0 )
		{
			RtlInsertToListTail(
				&m_TaskMgr.waitobj_list,
				&pDispatcherHeader->waitobj_list
				);
		}
		
		// register the current task to dispatcher wait list
		RtlInsertToListTail( &pDispatcherHeader->wait_list, &pt_task->wait_list );
		pt_task->wait_timeout = (Timeout==INFINITE ? INFINITE : KeQueryTickCount()+Timeout);

		// unregister the current task from the task ready list
		RtlRemoveEntryFromList( &pt_task->ready_list );

		// change the task status to waiting
		pt_task->status = TASK_STATUS_WAITING;
	}
	ENABLE_INTERRUPTS();
	
	// wait scheduler
	TmWaitScheduler();

	return TRUE;
}

KERNELAPI
KE_WAIT_STATUS
TmQueryWaitStatus(VOID)
{
	return m_TaskMgr.pt_cur_task->wait_status;
}

//////////////////////////////////////////////////////////////////////////////////////////
static
VOID
TmpReleaseTimeoutedWaitingTasks(
	IN	PLIST_ENTRY						pWaitList
	)
{
	ULONG			cur_tick;
	PLIST_ENTRY		pt_entry = NULL;
	
	// init
	cur_tick = KeQueryTickCount();
	
	// wakeup!
	for( pt_entry=pWaitList->Flink; pt_entry!=pWaitList; )
	{
		PTASK_CONTROL_BLOCK		pt_task = NULL;
		
		// get task block
		pt_task = PspGetTcbByWaitList( pt_entry );
		
		// get next entry!
		pt_entry = pt_entry->Flink;
		
		// is timeouted?
		if( pt_task->wait_timeout <= cur_tick )
		{
			TmpWakeupTask( pt_task, KE_WAIT_TIMEOUT );
		}
	}
}

static
VOID
TmpReleaseTasksWaitingObject(VOID)
{
	PLIST_ENTRY			pt_list = &m_TaskMgr.waitobj_list;
	PLIST_ENTRY			pt_entry = NULL;

	// wakeup loop!
	for( pt_entry=pt_list->Flink; pt_entry!=pt_list; )
	{
		KE_WAIT_STATUS		wait_result;
		PDISPATCHER_HEADER	pt_header = NULL;
		
		// get waitobj
		pt_header = DpcGetObjectByWaitObjList( pt_entry );
		
		// wakeup loop!
		while( pt_header->wait_list.Flink != &pt_header->wait_list )
		{
			// try object
			wait_result = KeWaitForSingleObject( (PVOID)pt_header, 0 );
			if( wait_result == KE_WAIT_TIMEOUT )
			{
				TmpReleaseTimeoutedWaitingTasks( &pt_header->wait_list );
				break;
			}
			
			// wakeup!
			TmWakeupFirstTaskWaitingObject(
				&pt_header->wait_list,
				KE_WAIT_SUCCESS
				);
		}
		if( pt_header->wait_list.Flink == &pt_header->wait_list )
		{
			pt_entry = pt_entry->Flink;
			RtlRemoveEntryFromList( pt_entry->Blink );
			continue;
		}
	
		// next waitobj
		pt_entry = pt_entry->Flink;
	}
}

KERNELAPI
VOID
TmIrqScheduler(
	IN	ULONG							StackCurrent
	)
{
	PTASK_CONTROL_BLOCK		pt_cur_task = NULL;
	PTASK_CONTROL_BLOCK		pt_next_task = NULL;
	ULONG					i;

	// init
	pt_cur_task = m_TaskMgr.pt_cur_task;

	// adjust current task's lr value
	pt_cur_task->stack_cur = StackCurrent;
	*((PULONG)pt_cur_task->stack_cur + 15) -= 4;

	// change the task status
	if( pt_cur_task->status == TASK_STATUS_RUNNING )
	{
		pt_cur_task->status = TASK_STATUS_READY;
	}
	// is terminated task? then, do remained works
	else if( pt_cur_task->status == TASK_STATUS_TERMINATED )
	{
		MmFree( (PVOID)pt_cur_task->stack_start );
	}

	// release tasks waiting for object
	TmpReleaseTasksWaitingObject();

	// update kernel tickcount
	KeUpdateTickCount( MSECS_PER_TMR_INT );

	// get next task
	for(i=0; i<PRIORITY_GROUP_COUNT; i++)
	{
		if( RtlGetListCount( &m_TaskMgr.ready_list[i] ) == 0 ) { continue; }

		pt_next_task = PspGetTcbByReadyList( m_TaskMgr.ready_list[i].Flink );
		RtlRemoveEntryFromList( &pt_next_task->ready_list );
		RtlInsertToListTail( &m_TaskMgr.ready_list[i], &pt_next_task->ready_list );
		break;
	}
	if( pt_next_task == NULL )
	{
		DbgPrint("task not found");
		while(1) ;
	}

	// change the new task's status
	pt_next_task->status = TASK_STATUS_RUNNING;

	// set new current task
	m_TaskMgr.pt_cur_task = pt_next_task;

	// clear interrupt
	CLEAR_TIMER_INTERRUPT_FLAG();

	// task switch!
	pt_next_task->stack_cur += CONTEXT_DATA_SIZE;
	TmpTaskContextSwitch(
		pt_next_task->stack_cur - CONTEXT_DATA_SIZE
		);
}

//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
HANDLE
TmGetCurrentTask(VOID)
{
	return (HANDLE)m_TaskMgr.pt_cur_task;
}

KERNELAPI
BOOL
TmSuspendTask(
	IN	HANDLE							TaskHandle
	)
{
	PTASK_CONTROL_BLOCK		pt_task = NULL;
	BOOL					is_running_task;

	// check parameter
	if( !TaskHandle ) { return FALSE; }

	// get block
	pt_task = PspGetTcbByHandle(TaskHandle);

	// check the task status
	if( pt_task->status != TASK_STATUS_READY &&
		pt_task->status != TASK_STATUS_RUNNING )
	{
		return FALSE;
	}
	
	// is running task?
	is_running_task = (pt_task->status == TASK_STATUS_RUNNING ? TRUE : FALSE);

	// change!
	DISABLE_INTERRUPTS();
	//
	RtlRemoveEntryFromList( &pt_task->ready_list );
	pt_task->status = TASK_STATUS_SUSPENDED;
	//
	ENABLE_INTERRUPTS();

	// wait scheduler
	if( is_running_task )
	{
		TmWaitScheduler();
	}

	return TRUE;
}

KERNELAPI
BOOL
TmResumeTask(
	IN	HANDLE							TaskHandle
	)
{
	PTASK_CONTROL_BLOCK		pt_task = NULL;

	// check parameter
	if( !TaskHandle ) { return FALSE; }

	// get block
	pt_task = PspGetTcbByHandle(TaskHandle);

	// check the task status
	if( pt_task->status == TASK_STATUS_TERMINATED )	{ return FALSE; }
	if( pt_task->status != TASK_STATUS_SUSPENDED )	{ return TRUE; }

	// change!
	DISABLE_INTERRUPTS();
	//
	RtlInsertToListHead(
		&m_TaskMgr.ready_list[ (ULONG)pt_task->priority ],
		&pt_task->ready_list
		);
	pt_task->status = TASK_STATUS_READY;
	//
	ENABLE_INTERRUPTS();
	
	// wait scheduler
	TmWaitScheduler();

	return TRUE;
}

KERNELAPI
BOOL
TmGetTaskPriority(
	IN	HANDLE							TaskHandle,
	OUT	PTASK_PRIORITY					pPriority
	)
{
	PTASK_CONTROL_BLOCK		pt_task = NULL;

	// check parameter
	if( !TaskHandle || !pPriority ) { return FALSE; }

	// get block
	pt_task = PspGetTcbByHandle(TaskHandle);

	// set
	*pPriority = pt_task->priority;

	return TRUE;
}

static
BOOL
TmpSetTaskPriority(
	IN	HANDLE							TaskHandle,
	IN	TASK_PRIORITY					Priority
	)
{
	PTASK_CONTROL_BLOCK		pt_task = NULL;

	// check parameter
	if( !TaskHandle ) { return FALSE; }

	// get block
	pt_task = PspGetTcbByHandle(TaskHandle);

	// is same priority? then, skip!
	if( pt_task->priority == Priority ) { return TRUE; }

	// change!
	DISABLE_INTERRUPTS();
	//
	pt_task->priority = Priority;
	if( pt_task->status == TASK_STATUS_READY || pt_task->status == TASK_STATUS_RUNNING )
	{
		RtlRemoveEntryFromList( &pt_task->ready_list );
		RtlInsertToListHead(
			&m_TaskMgr.ready_list[ (ULONG)Priority ],
			&pt_task->ready_list
			);
	}
	//
	ENABLE_INTERRUPTS();

	return TRUE;
}

KERNELAPI
BOOL
TmSetTaskPriority(
	IN	HANDLE							TaskHandle,
	IN	TASK_PRIORITY					Priority
	)
{
	if( !TmpSetTaskPriority(TaskHandle, Priority) )
	{
		return FALSE;
	}
	TmWaitScheduler();
	
	return TRUE;
}


//////////////////////////////////////////////////////////////////////////////////////////
KERNELAPI
BOOL
TmTerminateTask(
	IN	HANDLE							TaskHandle
	)
{
	PTASK_CONTROL_BLOCK		pt_task = NULL;
	BOOL					is_running_task;

	// check parameter
	if( !TaskHandle ) { return FALSE; }

	// get block
	pt_task = PspGetTcbByHandle(TaskHandle);
	
	// is not a task or already terminated?
	if( pt_task->header.signature != TCB_SIGNATURE ||
		pt_task->status == TASK_STATUS_TERMINATED )
	{
		return FALSE;
	}
	
	// is running task?
	is_running_task = (pt_task->status == TASK_STATUS_RUNNING ? TRUE : FALSE);
	
	// terminate!
	DISABLE_INTERRUPTS();
	{
		// Ÿ ½ũ  ڿ Locking ϰ ִ ¿ Ǹ
		//  Ʈ ٸ ִ Task   Ѵ.
		// ̸ ذϴ ڵ尡 ʿ....
		
		// remove from task & ready & wait list
		RtlRemoveEntryFromList( &pt_task->task_list );
		RtlRemoveEntryFromList( &pt_task->ready_list );
		RtlRemoveEntryFromList( &pt_task->wait_list );
		
		// mark as signaled state
		pt_task->header.signal_state = 1;
		TmWakeupAllTasksWaitingObject( &pt_task->header.wait_list, KE_WAIT_SUCCESS );
		
		// free stack buffer, if this task is not running
		if( !is_running_task )
		{
			MmFree( (PVOID)pt_task->stack_start );
		}
		
		// mark its status as terminated
		pt_task->status = TASK_STATUS_TERMINATED;
	}
	ENABLE_INTERRUPTS();
	
	// wait scheduler, if this task is running
	if( is_running_task )
	{
		TmWaitScheduler();
	}

	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////////////////
static
BOOL
TmpBuildTaskStackFrame(
	IN	PULONG							pStackFrame,
	IN	ULONG							StartRoutine,
	IN	ULONG							Parameter,
	IN	ULONG							CpsrValue
	)
// ---------------------
// Stack Frame Structure
// ---------------------
// 15 : lr (exactly, it means program counter(PC).)
// 14 : r12
// 13 : r3
// 12 : r2
// 11 : r1
// 10 : r0
// 09 : r11
// 08 : r10
// 07 : r9
// 06 : r8
// 05 : r7
// 04 : r6
// 03 : r5
// 02 : r4
// 01 : lr (r14)
// 00 : cpsr
{
	// check parameters
	if( !pStackFrame || !StartRoutine || !CpsrValue ) { return FALSE; }

	// build stack
	pStackFrame[0]	= CpsrValue;
	pStackFrame[1]	= (ULONG)TmpTaskSelfDestroy;
	pStackFrame[10]	= Parameter;
	pStackFrame[15]	= StartRoutine;

	return TRUE;
}

KERNELAPI
HANDLE
TmCreateTask(
	IN	ULONG							StackSize,		// 0 = default
	IN	PTASK_START_ROUTINE				pStartRoutine,
	IN	PVOID							pParameter,
	IN	BOOL							bSuspend
	)
{
	BOOL					result;
	PTASK_CONTROL_BLOCK		pt_task = NULL;
	PULONG					pt_stack = NULL;

	// check parameters
	if( !pStartRoutine ) { return NULL; }
	
	// adjust StackSize
	StackSize = (StackSize ? StackSize : DEFAULT_STACK_SIZE);
	
	// alloc mem
	pt_task		= (PTASK_CONTROL_BLOCK)MmAllocate( sizeof(TASK_CONTROL_BLOCK) );
	pt_stack	= (PULONG)MmAllocate( StackSize );
	if( !pt_task || !pt_stack ) { goto $error_occurred; }

	// init task control block
	memset( pt_task, 0, sizeof(TASK_CONTROL_BLOCK) );
	KeInitializeDispatcherHeader( &pt_task->header, TCB_SIGNATURE, 0 );
	RtlInitializeList( &pt_task->wait_list );
	pt_task->wait_timeout			= 0;
	pt_task->wait_status			= KE_WAIT_SUCCESS;
	pt_task->priority				= TASK_PRIORITY_NORMAL;
	pt_task->status					= (bSuspend ? TASK_STATUS_SUSPENDED : TASK_STATUS_READY);
	pt_task->stack_size				= StackSize;
	pt_task->stack_start			= (ULONG)pt_stack;
	pt_task->stack_cur
		= (pt_task->stack_start + pt_task->stack_size - CONTEXT_DATA_SIZE);

	// build stack frame
	result = TmpBuildTaskStackFrame(
		(PULONG)pt_task->stack_cur,
		(ULONG)pStartRoutine,
		(ULONG)pParameter,
		(0x40 /* fiq disable */ | GBA_MODE_USER) //GBA_MODE_SUPERVISOR ) // 
		);
	if( !result ) { goto $error_occurred; }

	// add!
	do
	{
		DISABLE_INTERRUPTS();
		//
		RtlInsertToListHead( &m_TaskMgr.task_list, &pt_task->task_list );
		if( bSuspend )
		{
			RtlInitializeList( &pt_task->ready_list );
		}
		else
		{
			RtlInsertToListHead(
				&m_TaskMgr.ready_list[TASK_PRIORITY_NORMAL], 
				&pt_task->ready_list
				);
		}
		pt_task->tid = m_TaskMgr.next_tid++;
		m_TaskMgr.task_count++;
		//
		ENABLE_INTERRUPTS();
	} while(FALSE);

	return (HANDLE)pt_task;

	//
$error_occurred:
	if( pt_stack )
	{
		MmFree( pt_stack ); pt_stack = NULL;
	}
	if( pt_task )
	{
		MmFree( pt_task ); pt_task = NULL;
	}

	return NULL;
}


//////////////////////////////////////////////////////////////////////////////////////////
static
VOID
IdleTask(PVOID pContext)
{
	while(1)
	{
		;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
static
VOID
TmpInitTaskManagerBlock(VOID)
{
	ULONG	i;

	//
	memset( &m_TaskMgr, 0, sizeof(m_TaskMgr) );
	m_TaskMgr.pt_cur_task = NULL;
	RtlInitializeList( &m_TaskMgr.task_list );
	RtlInitializeList( &m_TaskMgr.waitobj_list );
	for(i=0; i<PRIORITY_GROUP_COUNT; i++)
	{
		RtlInitializeList( &m_TaskMgr.ready_list[i] );
	}
}

KERNELAPI
BOOL
TmInitializeModule(VOID)
{
	// init
	TmpInitTaskManagerBlock();

	// create IDLE_TASK
	m_TaskMgr.pt_cur_task = (PTASK_CONTROL_BLOCK)TmCreateTask( 0, IdleTask, NULL, FALSE );
	TmpSetTaskPriority( (HANDLE)m_TaskMgr.pt_cur_task, TASK_PRIORITY_LOWEST );

	return TRUE;
}
